/*

                                   _JNJ`                        
                               .JNMH`                           
                              JMMF`       `;.                   
                            .NMM)          `MN.                 
                            MMM)            (MML                
                           (MMM`             MMML               
                     M  I  D  N  I  G  H  T    C  o  D  E       
                           (NMMF             MHNH               
                            NMML            .MMM                
                             NMML          .NMH                 
                              4MMNL       .#F                   
                               `4HNNL_   `                      
                                   `"""`                        

                          Copyright (C) 2004-2012
       "Ian (Larry) Latter" <ian dot latter at midnightcode dot org>

    Midnight Code is a registered trademark of Ian Latter.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, and mirrored at the Midnight Code web
    site; as at version 2 of the License only.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    (version 2) along with this program; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA, or see http://midnightcode.org/gplv2.txt

*/


/*
  Hardware;
  
    In Australia, JayCar stocks all of the needed components;
      9v Supply:   http://jaycar.com.au/productView.asp?ID=MP3146
      Patch lead:  http://jaycar.com.au/productView.asp?ID=YN8282
      Etherten:    http://jaycar.com.au/productView.asp?ID=XC4216
           Source: http://www.freetronics.com/products/etherten


  Software;
  
    Compiled via Arduino IDE
      arduino-022: http://arduino.cc/en/Main/Software
      
    With SD, this is a big sketch, so it errors on the Uno boot loader;
      Big seketch: http://arduino.cc/forum/index.php?topic=73207.0
      
    Requires the fixed Optiboot boot loader;
      Optiboot:    http://arduino.cc/forum/index.php/topic,64105.msg466703.html#msg466703
           Source: https://github.com/WestfW/Arduino
                   https://raw.github.com/WestfW/Arduino/master/hardware/arduino/bootloaders/optiboot/optiboot_atmega328.hex
                   https://raw.github.com/WestfW/Arduino/master/hardware/arduino/bootloaders/optiboot/optiboot_atmega328.lst
                
    Which can be installed from another Arduino, thanks to Optifix;
      Optifix:     http://www.adafruit.com/forums/viewtopic.php?f=25&t=19260
          Source:  https://docs.google.com/uc?id=0B6dMB5dovDUZM2ZhMmUxYTItM2ExNy00MWM1LWIzMTgtNWI0ZGUwYTNhZTE2&export=download&hl=en
      ArduinoISP:  http://arduino.cc/en/Tutorial/ArduinoISP
    
    Requires the following add-on/non-standard libraries;
      Ethernet:    http://gkaindl.com/software/arduino-ethernet/dhcp
      Time/Alarm:  http://arduino.cc/playground/Code/Time
      SDFAT:       http://code.google.com/p/sdfatlib/downloads/list

    Thanks also
      LadyAda:     http://www.ladyada.net/learn/arduino/ethfiles.html
      Alex2985:    http://www.youtube.com/watch?v=z96-Kfo87TQ
    
*/
#include <SdFat.h>
#include <SdFatUtil.h>
#include <SPI.h>
#include <Ethernet.h>
// #include <EthernetDHCP.h>
#include <Time.h>


// Custom Variables - your customisation should go here

// Features on/off
int ip_address_is_static  = true;          // true/false - do you want static addressing?
int web_server_is_enabled = true;          // true/false - do you want a web serving clock?

// Layer-4 addressing
unsigned int httpPort     = 80;

// Layer-3 addressing
byte static_ipAddr[]      = { 192, 168, 1, 18 };
byte static_gatewayAddr[] = { 192, 168, 1, 1 };
byte static_dnsAddr[]     = { 192, 168, 1, 1 };
byte static_ntpAddr[]     = { 192, 168, 1, 1 };

// Layer-2 addressing
byte mac[]                = { 0xC0, 0xDE, 0x00, 0x00, 0x00, 0x01 };



// Static Variables
// DHCP
const char* ip_to_str(const uint8_t*);
const byte* ipAddr;
const byte* gatewayAddr;
const byte* dnsAddr;
const byte* ntpAddr;
      
// HTTP
Server http_server(httpPort);
String http_default_file  = "index.htm";
unsigned int http_default_content_type_index = 4;
char * http_mime_types[]  = {".css", "text/css",
                             ".htm", "text/html",
                             ".js",  "application/javascript", 
                             ".png", "image/png",
                             ".txt", "text/plain"
                            };
#define http_mime_types_count (sizeof(http_mime_types)/sizeof(char *))
#define client_input_max 64

// SD
Sd2Card sd_card;
SdVolume sd_volume;
SdFile sd_root;
int sd_ss_pin = 10;
int sd_cs_pin = 4;
int sd_available = 0;



void ram_print() {
  Serial.print("Free RAM: ");
  Serial.println(FreeRam());  
  return;
}


void setup() {
  Serial.begin(9600);
  ram_print();
  
  sd_available = 0;
  pinMode(sd_ss_pin, OUTPUT);
  digitalWrite(sd_ss_pin, HIGH);
  if(sd_card.init(SPI_FULL_SPEED, sd_cs_pin)) {
    if(sd_volume.init(&sd_card)) {
      if(sd_root.openRoot(&sd_volume)) {
        sd_available = 1;
      }
    }
  }

  if(ip_address_is_static) {
    Ethernet.begin(mac, static_ipAddr, static_gatewayAddr);
    ipAddr = static_ipAddr;
    gatewayAddr = static_gatewayAddr;
    dnsAddr = static_dnsAddr;
/*  } else {
    EthernetDHCP.begin(mac);
    ipAddr = EthernetDHCP.ipAddress();
    gatewayAddr = EthernetDHCP.gatewayIpAddress();
    dnsAddr = EthernetDHCP.dnsIpAddress();
*/
  }
  
  if(web_server_is_enabled) {
    http_server.begin();
  }
  
  ram_print();
}



void loop() {
  if(!ip_address_is_static) {
    // EthernetDHCP.maintain();
  }
  if(web_server_is_enabled) {
    web_server();  
  }
}



int last_stamp(char * last_modified, uint16_t fatDate, uint16_t fatTime) {
  tmElements_t tm;
  tmElements_t * tm_p;
  unsigned long time_t;
  char day_c[4];
  char month_c[4];

  memset(&tm, 0, sizeof(tm));
  memset(last_modified, 0, 30);
  memset(&day_c, 0, 4);
  memset(&month_c, 0, 4);
 
  tm.Year   = FAT_YEAR(fatDate) - 1970;
  tm.Month  = FAT_MONTH(fatDate);
  tm.Day    = FAT_DAY(fatDate);
  tm.Hour   = FAT_HOUR(fatTime);
  tm.Minute = FAT_MINUTE(fatTime);
  tm.Second = FAT_SECOND(fatTime);
  
  time_t = makeTime(tm);

  if(snprintf(day_c, 4, "%.3s", dayShortStr(dayOfWeek(time_t))) < 0)
    return -1;
  if(snprintf(month_c, 4, "%.3s", monthShortStr(month(time_t))) < 0)
    return -1;
  if(snprintf(last_modified, 30, 
    "%s, %02d %s %04d %02d:%02d:%02d GMT",
    day_c,  day(time_t), month_c, year(time_t),
    hour(time_t), minute(time_t), second(time_t)) < 0)
    return -1;

  return 0;
}


  
void web_server() {
  char filename[13];
  char * filename_c;
  String filename_s;  
  unsigned int mime_index;
  unsigned int content_type_index;
  char last_modified[30];
  unsigned long file_size;
  int16_t file_byte;
  boolean client_input_empty;
  char client_input[client_input_max];
  int client_input_index;
  boolean client_error;
  dir_t dir;
  SdFile file;

   
  Client client = http_server.available();
  if (client) {
    client_input_empty = true;
    client_input_index = 0;
    client_error = false;
    while (client.connected()) {
      if (client.available()) {
        char c = client.read();
        // TODO: check for CRLFCRLF then get If-Modified-Since header
        if (c != '\n' && c != '\r') {
          client_input[client_input_index] = c;
          client_input_index++;
          if (client_input_index >= client_input_max) 
            client_input_index = client_input_max - 1;
          continue;
        }
        client_input[client_input_index] = 0;
        if (strstr(client_input, "HEAD /") != 0 || strstr(client_input, "GET /") != 0) {
          Serial.print("HEAD/GET: [");
          Serial.print(String(client_input));
          Serial.println("]");

          (strstr(client_input, " HTTP"))[0] = 0;   // replace " HTTP/1.1" with \0
          if(strstr(client_input, "GET /") != 0) {
            filename_c = client_input + 5;          // ignore the "GET /" (5 chars)
          } else {
            filename_c = client_input + 6;
          }
          filename_s = String(filename_c);
          if(filename_s == "") {
            filename_s = http_default_file;
            filename_c = filename;
            filename_s.toCharArray(filename_c, 13);
          }
          Serial.print("HEAD/GET: [");
          Serial.print(String(filename_c));
          Serial.println("]");

          if(file.open(&sd_root, filename_c, O_READ)) {
            file_size = file.fileSize();
            if(file.dirEntry(&dir)) {
              last_stamp((char *)&last_modified, dir.lastWriteDate, dir.lastWriteTime);
            }
            content_type_index = http_default_content_type_index;
            for(mime_index = 0; mime_index < http_mime_types_count - 1; mime_index += 2) {
              if(filename_s.lastIndexOf(http_mime_types[mime_index]) != -1) {
                content_type_index = mime_index + 1;
                break;
              }
            }
            ram_print();
            Serial.print("Sending response ..");
            client.println("HTTP/1.1 200 OK");
            client.print("Content-Type: ");
            client.println(http_mime_types[content_type_index]);
            if(last_modified != "") {
              client.print("Last-Modified: ");
              client.println(last_modified);
            }
            if(file_size > 0) {
              client.print("Content-Length: ");
              client.println(file_size);
            }
            client.println();
            if(strstr(client_input, "GET /") != 0) {
              Serial.print(".. and content ..");
              while ((file_byte = file.read()) > -1) {
                if(filename_s.equalsIgnoreCase("index.htm")) {
                  if(file_byte == 64) {
                    client.print("[Dynamic content goes here]");
                  } else {
                    client.print((char)file_byte);
                  }
                } else {
                  client.print((char)file_byte);
                }
              }
              // client.flush();
            }
            Serial.println(".. done");
            file.close();
            // delay(5);
            // client.stop();
          } else {
            client_error = true;
          }
        } else {
          client_error = true;
        }
        if(client_error) {
          client.println("HTTP/1.1 404 Not Found");
          client.println("Content-Type: text/html");
          client.println();
          client.println("<h2>File Not Found!</h2>");
          // client.flush();
          client_error = false;
          // delay(5);
          // client.stop();
        }
      }
    }
    // client.flush();
    delay(5);
    client.stop();
  }
}




